class PNGetSpaceMembershipsResult(object):
    def __init__(self, result):
        """
        Representation of get space memberships server response

        :param result: result of get space memberships operation
        """
        self.data = result['data']
        self.status = result['status']
        self.total_count = result.get('totalCount', None)
        self.next = result.get('next', None)
        self.prev = result.get('prev', None)

    def __str__(self):
        return "Get space memberships success with data: %s" % self.space


class PNManageMembershipsResult(object):
    def __init__(self, result):
        """
        Representation of manage memeberships response

        :param result: result of manage memeberships operation
        """
        self.data = result['data']
        self.status = result['status']
        self.total_count = result.get('totalCount', None)
        self.next = result.get('next', None)
        self.prev = result.get('prev', None)

    def __str__(self):
        return "Manage memeberships success with data: %s" % self.data


class PNGetMembersResult(object):
    def __init__(self, result):
        """
        Representation of fetch user server response

        :param result: result of fetch user operation
        """
        self.data = result['data']
        self.status = result['status']
        self.total_count = result.get('totalCount', None)
        self.next = result.get('next', None)
        self.prev = result.get('prev', None)

    def __str__(self):
        return "Get members success with data: %s" % self.data


class PNManageMembersResult(object):
    def __init__(self, result):
        """
        Representation of manage members server response

        :param result: result of manage members operation
        """
        self.data = result['data']
        self.status = result['status']
        self.total_count = result.get('totalCount', None)
        self.next = result.get('next', None)
        self.prev = result.get('prev', None)

    def __str__(self):
        return "Manage members success with data: %s" % self.data


class PNMembershipResult(object):
    def __init__(self, event, data):
        self.data = data
        self.event = event

    def __str__(self):
        return "Membership %s event with data: %s" % (self.event, self.data)



import six

from pubnub.endpoints.endpoint import Endpoint
from pubnub.managers import TokenManagerProperties
from pubnub.models.consumer.space import PNGetSpaceResult
from pubnub.enums import HttpMethod, PNOperationType, PNResourceType
from pubnub.exceptions import PubNubException


class GetSpace(Endpoint):
    GET_SPACE_PATH = '/v1/objects/%s/spaces/%s'

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._space_id = None
        self._include = None

    def space_id(self, space_id):
        assert isinstance(space_id, six.string_types)
        self._space_id = space_id
        return self

    def include(self, data):
        self._include = data
        return self

    def custom_params(self):
        params = {}
        if self._include:
            params['include'] = self._include
        return params

    def build_path(self):
        if self._space_id is None:
            raise PubNubException('Provide space id.')
        return GetSpace.GET_SPACE_PATH % (self.pubnub.config.subscribe_key, self._space_id)

    def http_method(self):
        return HttpMethod.GET

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()

    def create_response(self, envelope):   # pylint: disable=W0221
        return PNGetSpaceResult(envelope)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNGetSpaceOperation

    def name(self):
        return 'Get space'

    def get_tms_properties(self):
        return TokenManagerProperties(
            resource_type=PNResourceType.SPACE,
            resource_id=self._space_id if self._space_id is not None else ""
        )



import hashlib
import json
import sys

from .crypto_core import PubNubCrypto
from Cryptodome.Cipher import AES

Initial16bytes = '0123456789012345'

if sys.version_info > (3, 0):
    v = 3
else:
    v = 2

try:
    from base64 import decodebytes, encodebytes
except ImportError:
    from base64 import decodestring, encodestring

try:
    from hashlib import sha256
    digestmod = sha256
except ImportError:
    import Cryptodome.Hash.SHA256 as digestmod
    sha256 = digestmod.new


class PubNubCryptodome(PubNubCrypto):
    def encrypt(self, key, msg):
        secret = self.get_secret(key)

        if v == 3:
            cipher = AES.new(bytes(secret[0:32], 'utf-8'), AES.MODE_CBC, bytes(Initial16bytes, 'utf-8'))
            return encodebytes(cipher.encrypt(self.pad(msg.encode('utf-8')))).decode('utf-8').replace("\n", "")
        else:
            cipher = AES.new(secret[0:32], AES.MODE_CBC, Initial16bytes)
            return encodestring(cipher.encrypt(self.pad(msg))).replace("\n", "")

    def decrypt(self, key, msg):
        secret = self.get_secret(key)

        if v == 3:
            cipher = AES.new(bytes(secret[0:32], 'utf-8'), AES.MODE_CBC, bytes(Initial16bytes, 'utf-8'))
            plain = self.depad((cipher.decrypt(decodebytes(msg.encode('utf-8')))).decode('utf-8'))
        else:
            cipher = AES.new(secret[0:32], AES.MODE_CBC, Initial16bytes)
            plain = self.depad(cipher.decrypt(decodestring(msg)))

        try:
            return json.loads(plain)
        except Exception:
            return plain

    def pad(self, msg, block_size=16):
        padding = block_size - (len(msg) % block_size)

        if v == 3:
            return msg + (chr(padding) * padding).encode('utf-8')
        else:
            return msg + chr(padding) * padding

    def depad(self, msg):
        return msg[0:-ord(msg[-1])]

    def get_secret(self, key):
        if v == 3:
            return hashlib.sha256(key.encode("utf-8")).hexdigest()
        else:
            return hashlib.sha256(key).hexdigest()



import json
import logging
import time
import datetime

import math
import six
import tornado.gen
import tornado.httpclient
import tornado.ioloop
from tornado import gen

from tornado import ioloop
from tornado import stack_context
from tornado.concurrent import Future
from tornado.ioloop import PeriodicCallback
from tornado.locks import Event, Semaphore, Lock
from tornado.queues import Queue
from tornado.simple_httpclient import SimpleAsyncHTTPClient

from . import utils
from .models.consumer.common import PNStatus
from .callbacks import SubscribeCallback, ReconnectionCallback
from .endpoints.presence.leave import Leave
from .endpoints.pubsub.subscribe import Subscribe
from .enums import PNStatusCategory, PNHeartbeatNotificationOptions, PNOperationType, PNReconnectionPolicy
from .errors import PNERR_SERVER_ERROR, PNERR_CLIENT_ERROR, PNERR_JSON_DECODING_FAILED, PNERR_CLIENT_TIMEOUT, \
    PNERR_CONNECTION_ERROR
from .exceptions import PubNubException
from .managers import SubscriptionManager, PublishSequenceManager, ReconnectionManager, TelemetryManager
from .pubnub_core import PubNubCore
from .structures import ResponseInfo
from .workers import SubscribeMessageWorker

logger = logging.getLogger("pubnub")

tornado.httpclient.AsyncHTTPClient.configure(SimpleAsyncHTTPClient)


class PubNubTornado(PubNubCore):
    MAX_CLIENTS = 1000

    def stop(self):
        self.ioloop.stop()

    def start(self):
        self.ioloop.start()

    def timeout(self, delay, callback, *args):
        handle = None

        def cancel():
            self.ioloop.remove_timeout(handle)

        def cb():
            if callback is not None:
                callback(*args)

        handle = self.ioloop.add_timeout(time.time() + float(delay), cb)

        return cancel

    def sdk_platform(self):
        return "-Tornado"

    def __init__(self, config, custom_ioloop=None):
        super(PubNubTornado, self).__init__(config)
        self.ioloop = custom_ioloop or ioloop.IOLoop.instance()

        if self.config.enable_subscribe:
            self._subscription_manager = TornadoSubscriptionManager(self)

        self._publish_sequence_manager = TornadoPublishSequenceManager(PubNubCore.MAX_SEQUENCE)

        self.http = tornado.httpclient.AsyncHTTPClient(max_clients=PubNubTornado.MAX_CLIENTS)
        self.id = None

        self.headers = {
            'User-Agent': self.sdk_name,
            'Accept-Encoding': 'utf-8'
        }

        self._telemetry_manager = TornadoTelemetryManager(self.ioloop)

    def request_sync(self, *args):
        raise NotImplementedError

    def request_async(self, *args):
        raise NotImplementedError

    def request_deferred(self, *args):
        raise NotImplementedError

    @tornado.gen.coroutine
    def request_result(self, options_func, cancellation_event):
        try:
            envelope = yield self._request_helper(options_func, cancellation_event)
            raise tornado.gen.Return(envelope.result)
        except PubNubTornadoException as ex:
            raise ex.status.error_data.exception

    @tornado.gen.coroutine
    def request_future(self, options_func, cancellation_event):
        try:
            e = yield self._request_helper(options_func, cancellation_event)
        except PubNubTornadoException as ex:
            e = ex
        except Exception as ex:
            e = PubNubTornadoException(
                result=None,
                status=options_func().create_status(PNStatusCategory.PNUnknownCategory,
                                                    None,
                                                    None,
                                                    ex)
            )

        raise tornado.gen.Return(e)

    # REFACTOR: quickly adjusted to fit the new result() and future() endpoints
    def _request_helper(self, options_func, cancellation_event):
        if cancellation_event is not None:
            assert isinstance(cancellation_event, Event)

        # validate_params()
        options = options_func()

        create_response = options.create_response
        create_status_response = options.create_status

        params_to_merge_in = {}

        if options.operation_type == PNOperationType.PNPublishOperation:
            params_to_merge_in['seqn'] = self._publish_sequence_manager.get_next_sequence()

        options.merge_params_in(params_to_merge_in)

        future = Future()

        url = utils.build_url(self.config.scheme(), self.base_origin,
                              options.path, options.query_string)

        logger.debug("%s %s %s" % (options.method_string, url, options.data))

        if options.method_string == "POST":
            self.headers['Content-type'] = "application/json"

        start_timestamp = time.time()

        request = tornado.httpclient.HTTPRequest(
            url=url,
            method=options.method_string,
            headers=self.headers,
            body=options.data if options.data is not None else None,
            connect_timeout=options.connect_timeout,
            request_timeout=options.request_timeout)

        def response_callback(response):
            if cancellation_event is not None and cancellation_event.is_set():
                return

            body = response.body
            response_info = None
            status_category = PNStatusCategory.PNUnknownCategory
            if response is not None:
                request_url = six.moves.urllib.parse.urlparse(response.effective_url)
                query = six.moves.urllib.parse.parse_qs(request_url.query)
                uuid = None
                auth_key = None

                if 'uuid' in query and len(query['uuid']) > 0:
                    uuid = query['uuid'][0]

                if 'auth_key' in query and len(query['auth_key']) > 0:
                    auth_key = query['auth_key'][0]

                response_info = ResponseInfo(
                    status_code=response.code,
                    tls_enabled='https' == request_url.scheme,
                    origin=request_url.hostname,
                    uuid=uuid,
                    auth_key=auth_key,
                    client_request=response.request
                )

            if body is not None and len(body) > 0:
                try:
                    data = json.loads(body)
                except (ValueError, TypeError):
                    try:
                        data = json.loads(body.decode("utf-8"))
                    except ValueError:
                        tornado_result = PubNubTornadoException(
                            create_response(None),
                            create_status_response(status_category, response, response_info, PubNubException(
                                pn_error=PNERR_JSON_DECODING_FAILED,
                                errormsg='json decode error')
                            )
                        )
                        future.set_exception(tornado_result)
                        return
            else:
                data = "N/A"

            logger.debug(data)

            if response.error is not None:
                if response.code >= 500:
                    err = PNERR_SERVER_ERROR
                    data = str(response.error)
                else:
                    err = PNERR_CLIENT_ERROR

                e = PubNubException(
                    errormsg=data,
                    pn_error=err,
                    status_code=response.code,
                )

                if response.code == 403:
                    status_category = PNStatusCategory.PNAccessDeniedCategory

                if response.code == 400:
                    status_category = PNStatusCategory.PNBadRequestCategory

                if response.code == 599:
                    if 'HTTP 599: Timeout during request' == data:
                        status_category = PNStatusCategory.PNTimeoutCategory
                        e = PubNubException(
                            pn_error=PNERR_CLIENT_TIMEOUT,
                            errormsg=str(e)
                        )
                    elif 'HTTP 599: Stream closed' == data or\
                            'Name or service not known' in data or\
                            'Temporary failure in name resolution' in data:
                        status_category = PNStatusCategory.PNNetworkIssuesCategory
                        e = PubNubException(
                            pn_error=PNERR_CONNECTION_ERROR,
                            errormsg=str(e)
                        )
                        # TODO: add check for other status codes

                future.set_exception(PubNubTornadoException(
                    result=data,
                    status=create_status_response(status_category, data, response_info, e)
                ))
            else:
                self._telemetry_manager.store_latency(time.time() - start_timestamp, options.operation_type)

                future.set_result(TornadoEnvelope(
                    result=create_response(data),
                    status=create_status_response(
                        PNStatusCategory.PNAcknowledgmentCategory,
                        data,
                        response_info,
                        None)
                ))

        self.http.fetch(
            request=request,
            callback=response_callback
        )

        return future


class TornadoReconnectionManager(ReconnectionManager):
    def __init__(self, pubnub):
        self._cancelled_event = Event()
        super(TornadoReconnectionManager, self).__init__(pubnub)

    @gen.coroutine
    def _register_heartbeat_timer(self):
        self._cancelled_event.clear()

        while not self._cancelled_event.is_set():
            if self._pubnub.config.reconnect_policy == PNReconnectionPolicy.EXPONENTIAL:
                self._timer_interval = int(math.pow(2, self._connection_errors) - 1)
                if self._timer_interval > self.MAXEXPONENTIALBACKOFF:
                    self._timer_interval = self.MINEXPONENTIALBACKOFF
                    self._connection_errors = 1
                    logger.debug("timerInterval > MAXEXPONENTIALBACKOFF at: %s" % utils.datetime_now())
                elif self._timer_interval < 1:
                    self._timer_interval = self.MINEXPONENTIALBACKOFF
                logger.debug("timerInterval = %d at: %s" % (self._timer_interval, utils.datetime_now()))
            else:
                self._timer_interval = self.INTERVAL

            # >>> Wait given interval or cancel
            sleeper = tornado.gen.sleep(self._timer_interval)
            canceller = self._cancelled_event.wait()

            wi = tornado.gen.WaitIterator(canceller, sleeper)

            while not wi.done():
                try:
                    future = wi.next()
                    yield future
                except Exception as e:
                    # TODO: verify the error will not be eaten
                    logger.error(e)
                    raise
                else:
                    if wi.current_future == sleeper:
                        break
                    elif wi.current_future == canceller:
                        return
                    else:
                        raise Exception("unknown future raised")

            logger.debug("reconnect loop at: %s" % utils.datetime_now())

            # >>> Attempt to request /time/0 endpoint
            try:
                yield self._pubnub.time().result()
                self._connection_errors = 1
                self._callback.on_reconnect()
                logger.debug("reconnection manager stop due success time endpoint call: %s" % utils.datetime_now())
                break
            except Exception:
                if self._pubnub.config.reconnect_policy == PNReconnectionPolicy.EXPONENTIAL:
                    logger.debug("reconnect interval increment at: %s" % utils.datetime_now())
                    self._connection_errors += 1

    def start_polling(self):
        if self._pubnub.config.reconnect_policy == PNReconnectionPolicy.NONE:
            logger.warning("reconnection policy is disabled, please handle reconnection manually.")
            return

        self._pubnub.ioloop.spawn_callback(self._register_heartbeat_timer)

    def stop_polling(self):
        if self._cancelled_event is not None and not self._cancelled_event.is_set():
            self._cancelled_event.set()


class TornadoPublishSequenceManager(PublishSequenceManager):
    def __init__(self, provided_max_sequence):
        super(TornadoPublishSequenceManager, self).__init__(provided_max_sequence)
        self._lock = Lock()
        self._ioloop = ioloop

    def get_next_sequence(self):
        if self.max_sequence == self.next_sequence:
            self.next_sequence = 1
        else:
            self.next_sequence += 1

        return self.next_sequence


class TornadoSubscribeMessageWorker(SubscribeMessageWorker):
    @tornado.gen.coroutine
    def run(self):
        self._take_message()

    @tornado.gen.coroutine
    def _take_message(self):
        i = 0
        while not self._event.is_set():
            try:
                msg = yield self._queue.get(datetime.timedelta(seconds=1))
                if msg is not None:
                    self._process_incoming_payload(msg)
                self._queue.task_done()
            except tornado.gen.TimeoutError:
                i += 1
                continue


class TornadoSubscriptionManager(SubscriptionManager):
    def __init__(self, pubnub_instance):

        subscription_manager = self

        self._message_queue = Queue()
        self._consumer_event = Event()
        self._cancellation_event = Event()
        self._subscription_lock = Semaphore(1)
        # self._current_request_key_object = None
        self._heartbeat_periodic_callback = None
        self._reconnection_manager = TornadoReconnectionManager(pubnub_instance)

        super(TornadoSubscriptionManager, self).__init__(pubnub_instance)
        self._start_worker()

        class TornadoReconnectionCallback(ReconnectionCallback):
            def on_reconnect(self):
                subscription_manager.reconnect()

                pn_status = PNStatus()
                pn_status.category = PNStatusCategory.PNReconnectedCategory
                pn_status.error = False

                subscription_manager._subscription_status_announced = True
                subscription_manager._listener_manager.announce_status(pn_status)

        self._reconnection_listener = TornadoReconnectionCallback()
        self._reconnection_manager.set_reconnection_listener(self._reconnection_listener)

    def _set_consumer_event(self):
        self._consumer_event.set()

    def _message_queue_put(self, message):
        self._message_queue.put(message)

    def _start_worker(self):
        self._consumer = TornadoSubscribeMessageWorker(self._pubnub,
                                                       self._listener_manager,
                                                       self._message_queue,
                                                       self._consumer_event)
        run = stack_context.wrap(self._consumer.run)
        self._pubnub.ioloop.spawn_callback(run)

    def reconnect(self):
        self._should_stop = False
        self._pubnub.ioloop.spawn_callback(self._start_subscribe_loop)
        # self._register_heartbeat_timer()

    def disconnect(self):
        self._should_stop = True
        self._stop_heartbeat_timer()
        self._stop_subscribe_loop()

    @tornado.gen.coroutine
    def _start_subscribe_loop(self):
        self._stop_subscribe_loop()

        yield self._subscription_lock.acquire()

        self._cancellation_event.clear()

        combined_channels = self._subscription_state.prepare_channel_list(True)
        combined_groups = self._subscription_state.prepare_channel_group_list(True)

        if len(combined_channels) == 0 and len(combined_groups) == 0:
            return

        envelope_future = Subscribe(self._pubnub) \
            .channels(combined_channels).channel_groups(combined_groups) \
            .timetoken(self._timetoken).region(self._region) \
            .filter_expression(self._pubnub.config.filter_expression) \
            .cancellation_event(self._cancellation_event) \
            .future()

        canceller_future = self._cancellation_event.wait()

        wi = tornado.gen.WaitIterator(envelope_future, canceller_future)

        # iterates 2 times: one for result one for cancelled
        while not wi.done():
            try:
                result = yield wi.next()
            except Exception as e:
                # TODO: verify the error will not be eaten
                logger.error(e)
                raise
            else:
                if wi.current_future == envelope_future:
                    e = result
                elif wi.current_future == canceller_future:
                    return
                else:
                    raise Exception("Unexpected future resolved: %s" % str(wi.current_future))

                if e.is_error():
                    # 599 error doesn't works - tornado use this status code
                    # for a wide range of errors, for ex:
                    # HTTP Server Error (599): [Errno -2] Name or service not known
                    if e.status is not None and e.status.category == PNStatusCategory.PNTimeoutCategory:
                        self._pubnub.ioloop.spawn_callback(self._start_subscribe_loop)
                        return

                    logger.error("Exception in subscribe loop: %s" % str(e))

                    if e.status is not None and e.status.category == PNStatusCategory.PNAccessDeniedCategory:
                        e.status.operation = PNOperationType.PNUnsubscribeOperation

                    self._listener_manager.announce_status(e.status)

                    self._reconnection_manager.start_polling()
                    self.disconnect()
                    return
                else:
                    self._handle_endpoint_call(e.result, e.status)

                    self._pubnub.ioloop.spawn_callback(self._start_subscribe_loop)

            finally:
                self._cancellation_event.set()
                yield tornado.gen.moment
                self._subscription_lock.release()
                self._cancellation_event.clear()
                break

    def _stop_subscribe_loop(self):
        if self._cancellation_event is not None and not self._cancellation_event.is_set():
            self._cancellation_event.set()

    def _stop_heartbeat_timer(self):
        if self._heartbeat_periodic_callback is not None:
            self._heartbeat_periodic_callback.stop()

    def _register_heartbeat_timer(self):
        super(TornadoSubscriptionManager, self)._register_heartbeat_timer()
        self._heartbeat_periodic_callback = PeriodicCallback(
            stack_context.wrap(self._perform_heartbeat_loop),
            self._pubnub.config.heartbeat_interval * TornadoSubscriptionManager.HEARTBEAT_INTERVAL_MULTIPLIER,
            self._pubnub.ioloop)
        self._heartbeat_periodic_callback.start()

    @tornado.gen.coroutine
    def _perform_heartbeat_loop(self):
        if self._heartbeat_call is not None:
            # TODO: cancel call
            pass

        cancellation_event = Event()
        state_payload = self._subscription_state.state_payload()
        presence_channels = self._subscription_state.prepare_channel_list(False)
        presence_groups = self._subscription_state.prepare_channel_group_list(False)

        if len(presence_channels) == 0 and len(presence_groups) == 0:
            return

        try:
            envelope = yield self._pubnub.heartbeat() \
                .channels(presence_channels) \
                .channel_groups(presence_groups) \
                .state(state_payload) \
                .cancellation_event(cancellation_event) \
                .future()

            heartbeat_verbosity = self._pubnub.config.heartbeat_notification_options
            if envelope.status.is_error:
                if heartbeat_verbosity == PNHeartbeatNotificationOptions.ALL or \
                        heartbeat_verbosity == PNHeartbeatNotificationOptions.ALL:
                    self._listener_manager.announce_status(envelope.status)
            else:
                if heartbeat_verbosity == PNHeartbeatNotificationOptions.ALL:
                    self._listener_manager.announce_status(envelope.status)

        except PubNubTornadoException:
            pass
            # TODO: check correctness
            # if e.status is not None and e.status.category == PNStatusCategory.PNTimeoutCategory:
            #     self._start_subscribe_loop()
            # else:
            #     self._listener_manager.announce_status(e.status)
        except Exception as e:
            print(e)
        finally:
            cancellation_event.set()

    @tornado.gen.coroutine
    def _send_leave(self, unsubscribe_operation):
        envelope = yield Leave(self._pubnub) \
            .channels(unsubscribe_operation.channels) \
            .channel_groups(unsubscribe_operation.channel_groups).future()
        self._listener_manager.announce_status(envelope.status)


class TornadoEnvelope(object):
    def __init__(self, result, status):
        self.result = result
        self.status = status

    @staticmethod
    def is_error():
        return False


class PubNubTornadoException(Exception):
    def __init__(self, result, status):
        self.result = result
        self.status = status

    def __str__(self):
        return str(self.status.error_data.exception)

    @staticmethod
    def is_error():
        return True

    def value(self):
        return self.status.error_data.exception


class SubscribeListener(SubscribeCallback):
    def __init__(self):
        self.connected = False
        self.connected_event = Event()
        self.disconnected_event = Event()
        self.presence_queue = Queue()
        self.message_queue = Queue()
        self.error_queue = Queue()

    def status(self, pubnub, status):
        if utils.is_subscribed_event(status) and not self.connected_event.is_set():
            self.connected_event.set()
        elif utils.is_unsubscribed_event(status) and not self.disconnected_event.is_set():
            self.disconnected_event.set()
        elif status.is_error():
            self.error_queue.put_nowait(status.error_data.exception)

    def message(self, pubnub, message):
        self.message_queue.put(message)

    def presence(self, pubnub, presence):
        self.presence_queue.put(presence)

    @tornado.gen.coroutine
    def _wait_for(self, coro):
        error = self.error_queue.get()
        wi = tornado.gen.WaitIterator(coro, error)

        while not wi.done():
            result = yield wi.next()

            if wi.current_future == coro:
                raise gen.Return(result)
            elif wi.current_future == error:
                raise result
            else:
                raise Exception("Unexpected future resolved: %s" % str(wi.current_future))

    @tornado.gen.coroutine
    def wait_for_connect(self):
        if not self.connected_event.is_set():
            yield self._wait_for(self.connected_event.wait())
        else:
            raise Exception("instance is already connected")

    @tornado.gen.coroutine
    def wait_for_disconnect(self):
        if not self.disconnected_event.is_set():
            yield self._wait_for(self.disconnected_event.wait())
        else:
            raise Exception("instance is already disconnected")

    @tornado.gen.coroutine
    def wait_for_message_on(self, *channel_names):
        channel_names = list(channel_names)
        while True:
            try:  # NOQA
                env = yield self._wait_for(self.message_queue.get())
                if env.channel in channel_names:
                    raise tornado.gen.Return(env)
                else:
                    continue
            finally:
                self.message_queue.task_done()

    @tornado.gen.coroutine
    def wait_for_presence_on(self, *channel_names):
        channel_names = list(channel_names)
        while True:
            try:
                try:
                    env = yield self._wait_for(self.presence_queue.get())
                except:  # NOQA E722 pylint: disable=W0702
                    break
                if env.channel in channel_names:
                    raise tornado.gen.Return(env)
                else:
                    continue
            finally:
                self.presence_queue.task_done()


class TornadoTelemetryManager(TelemetryManager):  # pylint: disable=W0612
    def __init__(self, ioloop):
        TelemetryManager.__init__(self)
        self.ioloop = ioloop
        self._timer = PeriodicCallback(
            stack_context.wrap(self._start_clean_up_timer),
            self.CLEAN_UP_INTERVAL * self.CLEAN_UP_INTERVAL_MULTIPLIER,
            self.ioloop)
        self._timer.start()

    @tornado.gen.coroutine
    def _start_clean_up_timer(self):
        self.clean_up_telemetry_data()

    def _stop_clean_up_timer(self):
        self._timer.stop()



import six

from pubnub.endpoints.endpoint import Endpoint
from pubnub.managers import TokenManagerProperties
from pubnub.models.consumer.user import PNDeleteUserResult
from pubnub.enums import HttpMethod, PNOperationType, PNResourceType
from pubnub.exceptions import PubNubException


class DeleteUser(Endpoint):
    DELETE_USER_PATH = '/v1/objects/%s/users/%s'

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._user_id = None

    def user_id(self, user_id):
        assert isinstance(user_id, six.string_types)
        self._user_id = user_id
        return self

    def custom_params(self):
        return {}

    def build_data(self):
        return

    def build_path(self):
        if self._user_id is None:
            raise PubNubException('Provide user_id.')
        return DeleteUser.DELETE_USER_PATH % (self.pubnub.config.subscribe_key, self._user_id)

    def http_method(self):
        return HttpMethod.DELETE

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()

    def create_response(self, envelope):   # pylint: disable=W0221
        return PNDeleteUserResult(envelope)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNDeleteUserOperation

    def name(self):
        return 'Delete user'

    def get_tms_properties(self):
        return TokenManagerProperties(
            resource_type=PNResourceType.USER,
            resource_id=self._user_id if self._user_id is not None else ""
        )



class PNGetUsersResult(object):
    def __init__(self, result):
        """
        Representation of get users server response

        :param result: result of get users operation
        """
        self.data = result['data']
        self.status = result['status']
        self.total_count = result.get('totalCount', None)
        self.next = result.get('next', None)
        self.prev = result.get('prev', None)

    def __str__(self):
        return "Get users success with data: %s" % self.data


class PNCreateUserResult(object):
    def __init__(self, result):
        """
        Representation of create user server response

        :param result: result of create user operation
        """
        self.data = result['data']
        self.status = result['status']

    def __str__(self):
        return "User created with data: %s" % self.data


class PNGetUserResult(object):
    def __init__(self, result):
        """
        Representation of get user server response

        :param result: result of get user operation
        """
        self.data = result['data']
        self.status = result['status']

    def __str__(self):
        return "Get user success with data: %s" % self.data


class PNUpdateUserResult(object):
    def __init__(self, result):
        """
        Representation of update user server response

        :param result: result of update user operation
        """
        self.data = result['data']
        self.status = result['status']

    def __str__(self):
        return "Update user success with data: %s" % self.data


class PNDeleteUserResult(object):
    def __init__(self, result):
        """
        Representation of delete user server response

        :param result: result of delete user operation
        """
        self.data = result['data']
        self.status = result['status']

    def __str__(self):
        return "Delete user success with data: %s" % self.data


class PNUserResult(object):
    def __init__(self, event, data):
        self.data = data
        self.event = event

    def __str__(self):
        return "User %s event with data: %s" % (self.event, self.data)



import logging
from abc import abstractmethod, ABCMeta

import math
import time
import copy
import base64
from cbor2 import loads

from . import utils
from .enums import PNStatusCategory, PNReconnectionPolicy, PNOperationType, PNResourceType, PNMatchType
from .models.consumer.common import PNStatus
from .models.server.subscribe import SubscribeEnvelope
from .dtos import SubscribeOperation, UnsubscribeOperation
from .callbacks import SubscribeCallback, ReconnectionCallback
from .models.subscription_item import SubscriptionItem
from .errors import PNERR_INVALID_ACCESS_TOKEN
from .exceptions import PubNubException

logger = logging.getLogger("pubnub")


class PublishSequenceManager(object):
    def __init__(self, provided_max_sequence):
        self.max_sequence = provided_max_sequence
        self.next_sequence = 0

    @abstractmethod
    def get_next_sequence(self):
        if self.max_sequence == self.next_sequence:
            self.next_sequence = 1
        else:
            self.next_sequence += 1
        return self.next_sequence


class BasePathManager(object):
    MAX_SUBDOMAIN = 20
    DEFAULT_SUBDOMAIN = "pubsub"
    DEFAULT_BASE_PATH = "pubnub.com"

    def __init__(self, initial_config):
        self.config = initial_config
        self._current_subdomain = 1

    def get_base_path(self):
        if self.config.origin is not None:
            return self.config.origin
        # TODO: should CacheBusting be used?
        elif False:
            constructed_url = ("ps%s.%s" % (self._current_subdomain, BasePathManager.DEFAULT_BASE_PATH))

            if self._current_subdomain == BasePathManager.MAX_SUBDOMAIN:
                self._current_subdomain = 1
            else:
                self._current_subdomain += 1

            return constructed_url
        else:
            return "%s.%s" % (BasePathManager.DEFAULT_SUBDOMAIN, BasePathManager.DEFAULT_BASE_PATH)


class ReconnectionManager(object):
    INTERVAL = 3
    MINEXPONENTIALBACKOFF = 1
    MAXEXPONENTIALBACKOFF = 32

    def __init__(self, pubnub):
        self._pubnub = pubnub
        self._callback = None
        self._timer = None
        self._timer_interval = None
        self._connection_errors = 1

    def set_reconnection_listener(self, reconnection_callback):
        assert isinstance(reconnection_callback, ReconnectionCallback)
        self._callback = reconnection_callback

    def _recalculate_interval(self):
        if self._pubnub.config.reconnect_policy == PNReconnectionPolicy.EXPONENTIAL:
            self._timer_interval = int(math.pow(2, self._connection_errors) - 1)
            if self._timer_interval > self.MAXEXPONENTIALBACKOFF:
                self._timer_interval = self.MINEXPONENTIALBACKOFF
                self._connection_errors = 1
                logger.debug("timerInterval > MAXEXPONENTIALBACKOFF at: %s" % utils.datetime_now())
            elif self._timer_interval < 1:
                self._timer_interval = self.MINEXPONENTIALBACKOFF
            logger.debug("timerInterval = %d at: %s" % (self._timer_interval, utils.datetime_now()))
        else:
            self._timer_interval = self.INTERVAL

    @abstractmethod
    def start_polling(self):
        pass

    def _stop_heartbeat_timer(self):
        if self._timer is not None:
            self._timer.stop()
            self._timer = None


class StateManager(object):
    def __init__(self):
        self._channels = {}
        self._groups = {}
        self._presence_channels = {}
        self._presence_groups = {}

    def is_empty(self):
        return len(self._channels) == 0 and len(self._groups) == 0 and\
            len(self._presence_channels) == 0 and len(self._presence_groups) == 0

    def subscribed_to_the_only_channel(self):
        return len(self._channels) == 1 and len(self._groups) == 0 and\
            len(self._presence_channels) == 0 and len(self._presence_groups) == 0

    def prepare_channel_list(self, include_presence):
        return StateManager._prepare_membership_list(
            self._channels, self._presence_channels, include_presence)

    def prepare_channel_group_list(self, include_presence):
        return StateManager._prepare_membership_list(
            self._groups, self._presence_groups, include_presence)

    def adapt_subscribe_builder(self, subscribe_operation):
        for channel in subscribe_operation.channels:
            self._channels[channel] = SubscriptionItem(name=channel)

            if subscribe_operation.presence_enabled:
                self._presence_channels[channel] = SubscriptionItem(name=channel)

        for group in subscribe_operation.channel_groups:
            self._groups[group] = SubscriptionItem(name=group)

            if subscribe_operation.presence_enabled:
                self._presence_groups[group] = SubscriptionItem(name=group)

    def adapt_unsubscribe_builder(self, unsubscribe_operation):
        for channel in unsubscribe_operation.channels:
            self._channels.pop(channel, None)
            if channel in self._presence_channels:
                self._presence_channels.pop(channel, None)

        for group in unsubscribe_operation.channel_groups:
            self._groups.pop(group)
            if group in self._presence_groups:
                self._presence_groups.pop(group)

    def adapt_state_builder(self, state_operation):
        for channel in state_operation.channels:
            subscribed_channel = self._channels.get(channel)

            if subscribed_channel is not None:
                subscribed_channel.state = state_operation.state

        for group in state_operation.channel_groups:
            subscribed_group = self._channels.get(group)

            if subscribed_group is not None:
                subscribed_group.state = state_operation.state

    def state_payload(self):
        state = {}

        for channel in self._channels.values():
            if channel.state is not None:
                state[channel.name] = channel.state

        for group in self._groups.values():
            if group.state is not None:
                state[group.name] = group.state

        return state

    @staticmethod
    def _prepare_membership_list(data_storage, presence_storage, include_presence):
        response = []

        for item in data_storage.values():
            response.append(item.name)

        if include_presence:
            for item in presence_storage.values():
                response.append(item.name + "-pnpres")

        return response


class ListenerManager(object):
    def __init__(self, pubnub_instance):
        self._pubnub = pubnub_instance
        self._listeners = []

    def add_listener(self, listener):
        assert isinstance(listener, SubscribeCallback)
        self._listeners.append(listener)

    def remove_listener(self, listener):
        assert isinstance(listener, SubscribeCallback)
        self._listeners.remove(listener)

    def announce_status(self, status):
        for callback in self._listeners:
            callback.status(self._pubnub, status)

    def announce_message(self, message):
        for callback in self._listeners:
            callback.message(self._pubnub, message)

    def announce_signal(self, signal):
        for callback in self._listeners:
            callback.signal(self._pubnub, signal)

    def announce_user(self, user):
        for callback in self._listeners:
            callback.user(self._pubnub, user)

    def announce_space(self, space):
        for callback in self._listeners:
            callback.space(self._pubnub, space)

    def announce_membership(self, membership):
        for callback in self._listeners:
            callback.membership(self._pubnub, membership)

    def announce_message_action(self, message_action):
        for callback in self._listeners:
            callback.message_action(self._pubnub, message_action)

    def announce_presence(self, presence):
        for callback in self._listeners:
            callback.presence(self._pubnub, presence)


class SubscriptionManager(object):
    __metaclass__ = ABCMeta

    HEARTBEAT_INTERVAL_MULTIPLIER = 1000

    def __init__(self, pubnub_instance):
        self._pubnub = pubnub_instance
        self._subscription_status_announced = False

        self._subscription_state = StateManager()
        self._listener_manager = ListenerManager(self._pubnub)
        self._timetoken = int(0)
        self._region = None

        self._should_stop = False

        self._subscribe_request_task = None
        self._heartbeat_call = None

    @abstractmethod
    def _start_worker(self):
        pass

    @abstractmethod
    def _set_consumer_event(self):
        pass

    @abstractmethod
    def _message_queue_put(self, message):
        pass

    @abstractmethod
    def _start_subscribe_loop(self):
        pass

    @abstractmethod
    def _stop_subscribe_loop(self):
        pass

    @abstractmethod
    def _stop_heartbeat_timer(self):
        pass

    @abstractmethod
    def _perform_heartbeat_loop(self):
        pass

    @abstractmethod
    def _send_leave(self, unsubscribe_operation):
        pass

    def add_listener(self, listener):
        self._listener_manager.add_listener(listener)

    def remove_listener(self, listener):
        self._listener_manager.remove_listener(listener)

    def get_subscribed_channels(self):
        return self._subscription_state.prepare_channel_list(False)

    def get_subscribed_channel_groups(self):
        return self._subscription_state.prepare_channel_group_list(False)

    def unsubscribe_all(self):
        self.adapt_unsubscribe_builder(UnsubscribeOperation(
            channels=self._subscription_state.prepare_channel_list(False),
            channel_groups=self._subscription_state.prepare_channel_group_list(False)
        ))

    def adapt_subscribe_builder(self, subscribe_operation):
        assert isinstance(subscribe_operation, SubscribeOperation)
        self._subscription_state.adapt_subscribe_builder(subscribe_operation)
        self._subscription_status_announced = False

        if subscribe_operation.timetoken is not None:
            self._timetoken = subscribe_operation.timetoken

        self.reconnect()

    def adapt_unsubscribe_builder(self, unsubscribe_operation):
        assert isinstance(unsubscribe_operation, UnsubscribeOperation)

        self._subscription_state.adapt_unsubscribe_builder(unsubscribe_operation)

        self._send_leave(unsubscribe_operation)

        if self._subscription_state.is_empty():
            self._region = None
            self._timetoken = 0
        self.reconnect()

    def adapt_state_builder(self, state_operation):
        self._subscription_state.adapt_state_builder(state_operation)
        self.reconnect()

    @abstractmethod
    def reconnect(self):
        pass

    def stop(self):
        self._should_stop = True
        self._stop_subscribe_loop()
        self._stop_heartbeat_timer()
        self._set_consumer_event()

    def _handle_endpoint_call(self, raw_result, status):
        assert isinstance(status, PNStatus)

        if not self._subscription_status_announced:
            pn_status = PNStatus()
            pn_status.category = PNStatusCategory.PNConnectedCategory
            pn_status.status_code = status.status_code
            pn_status.auth_key = status.auth_key
            pn_status.operation = status.operation
            pn_status.client_request = status.client_request
            pn_status.origin = status.origin
            pn_status.tls_enabled = status.tls_enabled

            self._subscription_status_announced = True
            self._listener_manager.announce_status(pn_status)

        result = SubscribeEnvelope.from_json(raw_result)
        only_channel = self._subscription_state.subscribed_to_the_only_channel()
        if result.messages is not None and len(result.messages) > 0:
            for message in result.messages:
                if only_channel:
                    message.only_channel_subscription = True
                self._message_queue_put(message)

        # REVIEW: is int compatible with long for Python 2
        self._timetoken = int(result.metadata.timetoken)
        self._region = int(result.metadata.region)

    # TODO: make abstract
    def _register_heartbeat_timer(self):
        self._stop_heartbeat_timer()


class TelemetryManager(object):  # pylint: disable=W0612
    TIMESTAMP_DIVIDER = 1000
    MAXIMUM_LATENCY_DATA_AGE = 60
    CLEAN_UP_INTERVAL = 1
    CLEAN_UP_INTERVAL_MULTIPLIER = 1000

    def __init__(self):
        self.latencies = {}

    @abstractmethod
    def _start_clean_up_timer(self):
        pass

    @abstractmethod
    def _stop_clean_up_timer(self):
        pass

    def operation_latencies(self):
        operation_latencies = {}

        for endpoint_name, endpoint_latencies in self.latencies.items():
            latency_key = 'l_' + endpoint_name

            endpoint_average_latency = self.average_latency_from_data(endpoint_latencies)

            if (endpoint_average_latency > 0):
                operation_latencies[latency_key] = endpoint_average_latency

        return operation_latencies

    def clean_up_telemetry_data(self):
        current_timestamp = time.time()

        copy_latencies = copy.deepcopy(dict(self.latencies))

        for endpoint_name, endpoint_latencies in copy_latencies.items():
            for latency_information in list(endpoint_latencies):
                if current_timestamp - latency_information['d'] > self.MAXIMUM_LATENCY_DATA_AGE:
                    self.latencies[endpoint_name].remove(latency_information)

            if len(self.latencies[endpoint_name]) == 0:
                del self.latencies[endpoint_name]

    def store_latency(self, latency, operation_type):
        if operation_type != PNOperationType.PNSubscribeOperation and latency > 0:
            endpoint_name = self.endpoint_name_for_operation(operation_type)

            store_timestamp = time.time()

            if endpoint_name not in self.latencies:
                self.latencies[endpoint_name] = []

            latency_entry = {
                'd': store_timestamp,
                'l': latency,
            }

            self.latencies[endpoint_name].append(latency_entry)

    @staticmethod
    def average_latency_from_data(endpoint_latencies):
        total_latency = 0

        for k in endpoint_latencies:
            total_latency += k['l']

        return total_latency / len(endpoint_latencies)

    @staticmethod
    def endpoint_name_for_operation(operation_type):
        endpoint = {
            PNOperationType.PNPublishOperation: 'pub',
            PNOperationType.PNFireOperation: 'pub',

            PNOperationType.PNHistoryOperation: 'hist',
            PNOperationType.PNHistoryDeleteOperation: 'hist',
            PNOperationType.PNMessageCountOperation: 'mc',

            PNOperationType.PNUnsubscribeOperation: 'pres',
            PNOperationType.PNWhereNowOperation: 'pres',
            PNOperationType.PNHereNowOperation: 'pres',
            PNOperationType.PNGetState: 'pres',
            PNOperationType.PNSetStateOperation: 'pres',

            PNOperationType.PNAddChannelsToGroupOperation: 'cg',
            PNOperationType.PNRemoveChannelsFromGroupOperation: 'cg',
            PNOperationType.PNChannelGroupsOperation: 'cg',
            PNOperationType.PNChannelsForGroupOperation: 'cg',
            PNOperationType.PNRemoveGroupOperation: 'cg',

            PNOperationType.PNAddPushNotificationsOnChannelsOperation: 'push',
            PNOperationType.PNPushNotificationEnabledChannelsOperation: 'push',
            PNOperationType.PNRemoveAllPushNotificationsOperation: 'push',
            PNOperationType.PNRemovePushNotificationsFromChannelsOperation: 'push',

            PNOperationType.PNAccessManagerAudit: 'pam',
            PNOperationType.PNAccessManagerGrant: 'pam',
            PNOperationType.PNAccessManagerRevoke: 'pam',
            PNOperationType.PNTimeOperation: 'pam',

            PNOperationType.PNSignalOperation: 'sig',

            PNOperationType.PNGetUsersOperation: 'obj',
            PNOperationType.PNCreateUserOperation: 'obj',
            PNOperationType.PNGetUserOperation: 'obj',
            PNOperationType.PNUpdateUserOperation: 'obj',
            PNOperationType.PNDeleteUserOperation: 'obj',
            PNOperationType.PNGetSpacesOperation: 'obj',
            PNOperationType.PNCreateSpaceOperation: 'obj',
            PNOperationType.PNGetSpaceOperation: 'obj',
            PNOperationType.PNUpdateSpaceOperation: 'obj',
            PNOperationType.PNDeleteSpaceOperation: 'obj',
            PNOperationType.PNGetMembersOperation: 'obj',
            PNOperationType.PNGetSpaceMembershipsOperation: 'obj',
            PNOperationType.PNManageMembersOperation: 'obj',
            PNOperationType.PNManageMembershipsOperation: 'obj',

            PNOperationType.PNAccessManagerGrantToken: 'pamv3',

            PNOperationType.PNAddMessageAction: 'msga',
            PNOperationType.PNGetMessageActions: 'msga',
            PNOperationType.PNDeleteMessageAction: 'msga'
        }[operation_type]

        return endpoint


class TokenManager(object):

    def __init__(self):
        self._map = {}
        self.init_map()

    def init_map(self):
        resources = [PNResourceType.USER, PNResourceType.SPACE]

        for resource in resources:
            skeleton_map = {
                PNMatchType.RESOURCE: {},
                PNMatchType.PATTERN: {}
            }
            self._map[resource] = skeleton_map

    def set_token(self, token):
        unwrapped_token = self.unwrap_token(token)
        self.store_token(unwrapped_token, token)

    def set_tokens(self, tokens):
        for token in tokens:
            self.set_token(token)

    def get_token(self, tms_properties):
        resource_token = self.get_token_by_match(tms_properties, PNMatchType.RESOURCE)

        if resource_token is None:
            return self.get_token_by_match(tms_properties, PNMatchType.PATTERN)

        return resource_token

    def get_tokens(self):
        return self._map

    def get_tokens_by_resource(self, resource_type):
        return self._map[resource_type]

    def store_token(self, unwrapped_token, token):
        match_types = [
            PNMatchType.RESOURCE,
            PNMatchType.PATTERN
        ]

        for asset in match_types:
            short_match_type = self.get_shortened_match_type(asset)

            if short_match_type in unwrapped_token:
                res_object = unwrapped_token[short_match_type]

                for r_type in res_object.keys():
                    single_res_object = res_object[r_type]
                    for r_name in single_res_object.keys():
                        if asset == PNMatchType.PATTERN:
                            self._map[self.get_extended_resource_type(r_type)][asset].clear()

                        self._map[self.get_extended_resource_type(r_type)][asset][r_name] = token

    def unwrap_token(self, token):
        raw = token

        raw = raw.replace("_", "/").replace("-", "+")
        byte_array = base64.b64decode(raw)

        try:
            unwrapped_obj = loads(byte_array)
            decoded_obj = utils.decode_utf8_dict(unwrapped_obj)

            return decoded_obj
        except Exception:
            raise PubNubException(pn_error=PNERR_INVALID_ACCESS_TOKEN)

    def get_token_by_match(self, tms_properties, match_type):
        if tms_properties is None or tms_properties.resource_type is None or tms_properties.resource_id is None:
            return None

        if match_type != PNMatchType.PATTERN:
            if tms_properties.resource_id in self._map[tms_properties.resource_type][match_type]:
                token = self._map[tms_properties.resource_type][match_type][tms_properties.resource_id]
                if token is not None:
                    return token
        else:
            string_token_wrapper_dict = self._map[tms_properties.resource_type][match_type]
            if len(string_token_wrapper_dict.keys()) > 0:
                first_key = list(string_token_wrapper_dict.keys())[0]
                return string_token_wrapper_dict[first_key]

        return None

    def get_extended_resource_type(self, r_type_abbr):
        if r_type_abbr == "usr":
            return PNResourceType.USER
        if r_type_abbr == "spc":
            return PNResourceType.SPACE

        return r_type_abbr

    def get_shortened_match_type(self, match_type):
        if match_type == PNMatchType.RESOURCE:
            return "res"
        if match_type == PNMatchType.PATTERN:
            return "pat"

        return match_type


class TokenManagerProperties:
    def __init__(self, resource_type, resource_id):
        self.resource_type = resource_type
        self.resource_id = resource_id

    def __str__(self):
        return "resource_type: " + self.resource_type + ", resource_id: " + self.resource_id



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.managers import TokenManagerProperties
from pubnub.models.consumer.space import PNCreateSpaceResult
from pubnub.enums import HttpMethod, PNOperationType, PNResourceType
from pubnub.exceptions import PubNubException


class CreateSpace(Endpoint):
    CREATE_SPACE_PATH = '/v1/objects/%s/spaces'

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._include = {}
        self._data = None

    def include(self, data):
        self._include = data
        return self

    def data(self, data):
        assert isinstance(data, dict)
        if 'id' not in data or 'name' not in data:
            raise PubNubException("Space's id or name missing.")
        self._data = data
        return self

    def custom_params(self):
        params = {}
        if self._include:
            params['include'] = self._include
        return params

    def build_data(self):
        return utils.write_value_as_string(self._data)

    def validate_params(self):
        self.validate_subscribe_key()
        if self._data is None:
            raise PubNubException('No data supplied.')

    def build_path(self):
        return CreateSpace.CREATE_SPACE_PATH % (self.pubnub.config.subscribe_key)

    def http_method(self):
        return HttpMethod.POST

    def is_auth_required(self):
        return True

    def create_response(self, envelope):  # pylint: disable=W0221
        return PNCreateSpaceResult(envelope)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNCreateSpaceOperation

    def name(self):
        return 'Create space'

    def get_tms_properties(self):
        return TokenManagerProperties(
            resource_type=PNResourceType.SPACE,
            resource_id=self._data['id'] if self._data is not None else ""
        )



import json
import logging
import time

from urlparse import urlparse, parse_qs
from StringIO import StringIO

from twisted.internet import reactor as _reactor
from twisted.internet.task import LoopingCall
from twisted.internet.defer import Deferred, DeferredQueue
from twisted.internet.protocol import Protocol
from twisted.internet.error import ConnectingCancelledError
from twisted.web.client import Agent
from twisted.web.client import HTTPConnectionPool
from twisted.web.http_headers import Headers
from twisted.web.client import FileBodyProducer

from . import utils
from .workers import SubscribeMessageWorker
from .pubnub_core import PubNubCore
from .managers import SubscriptionManager, PublishSequenceManager
from .enums import PNStatusCategory, PNHeartbeatNotificationOptions, PNOperationType
from .errors import PNERR_CLIENT_ERROR, PNERR_CONNECTION_ERROR, \
    PNERR_SERVER_ERROR, PNERR_JSON_DECODING_FAILED
from .exceptions import PubNubException
from .structures import ResponseInfo

from .endpoints.pubsub.subscribe import Subscribe
from .endpoints.presence.leave import Leave
from .endpoints.presence.heartbeat import Heartbeat

logger = logging.getLogger("pubnub")


class PubNubResponse(Protocol):
    def __init__(self, finished, code):
        self.finished = finished
        self.code = code

    def dataReceived(self, body):
        self.finished.callback(TwistedResponse(body, self.code))


class TwistedSubscribeMessageWorker(SubscribeMessageWorker):
    def run(self):
        self._take_message()

    def _take_message(self):
        self._queue.get().addCallback(self.send_message_to_processing)

    def send_message_to_processing(self, message):
        if message is not None:
            self._pubnub.reactor.callInThread(self._process_incoming_payload, message)


class TwistedSubscriptionManager(SubscriptionManager):
    def __init__(self, pubnub_instance):
        self._message_queue = DeferredQueue()
        self.worker_loop = None
        self._heartbeat_loop = None
        self._heartbeat_call = None
        self.clock = pubnub_instance.clock
        super(TwistedSubscriptionManager, self).__init__(pubnub_instance)

    def _announce_status(self, status):
        self._listener_manager.announce_status(status)

    def _start_worker(self):
        consumer = TwistedSubscribeMessageWorker(self._pubnub, self._listener_manager, self._message_queue, None)
        looping_call = LoopingCall(consumer.run)

        if self.clock is not None:
            looping_call.clock = self.clock

        self.worker_loop = looping_call.start(0.1, False)

    def _set_consumer_event(self):
        raise NotImplementedError

    def _message_queue_put(self, message):
        self._message_queue.put(message)

    def _start_subscribe_loop(self):
        self._stop_subscribe_loop()

        combined_channels = self._subscription_state.prepare_channel_list(True)
        combined_groups = self._subscription_state.prepare_channel_group_list(True)

        if len(combined_channels) == 0 and len(combined_groups) == 0:
            return

        def continue_subscribe_loop(envelope):
            try:
                self._handle_endpoint_call(envelope.raw_result, envelope.status)
            except Exception as ex:
                return ex
            self._start_subscribe_loop()

        def manage_failure(failure):
            if failure.type == PubNubTwistedException:
                self._announce_status(failure.value.status)
                if failure.value.status.category in (PNStatusCategory.PNDisconnectedCategory,
                                                     PNStatusCategory.PNUnexpectedDisconnectCategory,
                                                     PNStatusCategory.PNCancelledCategory,
                                                     PNStatusCategory.PNBadRequestCategory,
                                                     PNStatusCategory.PNMalformedFilterExpressionCategory):
                    time.sleep(30)  # TODO: SET VALUE ACCORDING TO DOCS
                    self._start_subscribe_loop()
            else:
                return failure

        try:
            self._subscribe_request_task = Subscribe(self._pubnub) \
                .channels(combined_channels) \
                .channel_groups(combined_groups) \
                .timetoken(self._timetoken) \
                .region(self._region) \
                .filter_expression(self._pubnub.config.filter_expression) \
                .deferred() \
                .addCallbacks(continue_subscribe_loop, manage_failure)

        except Exception as ex:
            raise ex

    def _stop_subscribe_loop(self):
        if self._subscribe_request_task is not None and not self._subscribe_request_task.called:
            self._subscribe_request_task.cancel()

    def _stop_heartbeat_timer(self):
        if self._heartbeat_call is not None:
            self._heartbeat_call.cancel()

        if self._heartbeat_loop is not None:
            self._heartbeat_loop.stop()

    def _register_heartbeat_timer(self):
        super(TwistedSubscriptionManager, self)._register_heartbeat_timer()
        self._heartbeat_loop = LoopingCall(self._perform_heartbeat_loop)
        interval = self._pubnub.config.heartbeat_interval / 2 - 1
        self._heartbeat_loop.start(interval, True)

    def _perform_heartbeat_loop(self):
        def heartbeat_callback(_, status):
            heartbeat_verbosity = self._pubnub.config.heartbeat_notification_options
            if heartbeat_verbosity == PNHeartbeatNotificationOptions.ALL or (
               status.is_error() is True and heartbeat_verbosity == PNHeartbeatNotificationOptions.FAILURES):
                self._listener_manager.announce_status(status)

        if self._heartbeat_call is not None:
            self._heartbeat_call.cancel()

        state_payload = self._subscription_state.state_payload()
        channels = self._subscription_state.prepare_channel_list(False)
        channel_groups = self._subscription_state.prepare_channel_group_list(False)

        self._heartbeat_call = Heartbeat(self._pubnub) \
            .channels(channels) \
            .channel_groups(channel_groups) \
            .state(state_payload) \
            .pn_async(heartbeat_callback)

    def _send_leave(self, unsubscribe_operation):
        def announce_leave_status(response, status):
            self._listener_manager.announce_status(status)

        Leave(self._pubnub) \
            .channels(unsubscribe_operation.channels) \
            .channel_groups(unsubscribe_operation.channel_groups) \
            .pn_async(announce_leave_status)

    def reconnect(self):
        # TODO: REVIEW
        self._start_subscribe_loop()
        self._register_heartbeat_timer()


class PubNubTwisted(PubNubCore):
    """PubNub Python API for Twisted framework"""

    def sdk_platform(self):
        return "-Twisted"

    def __init__(self, config, pool=None, reactor=None, clock=None):
        super(PubNubTwisted, self).__init__(config)

        self.clock = clock
        self._publish_sequence_manager = PublishSequenceManager(PubNubCore.MAX_SEQUENCE)

        if self.config.enable_subscribe:
            self._subscription_manager = TwistedSubscriptionManager(self)

        self.disconnected_times = 0

        if reactor is None:
            self.reactor = _reactor
        else:
            self.reactor = reactor

        if pool is None:
            self.pnconn_pool = HTTPConnectionPool(self.reactor, persistent=True)
            self.pnconn_pool.maxPersistentPerHost = 3
            self.pnconn_pool.cachedConnectionTimeout = self.config.subscribe_request_timeout
            self.pnconn_pool.retryAutomatically = False
        else:
            self.pnconn_pool = pool

        self.headers = {
            'User-Agent': [self.sdk_name],
        }

    def start(self, skip_reactor=False):
        if self._subscription_manager is not None:
            self._subscription_manager._start_worker()
        if not skip_reactor:
            self.reactor.run()

    def stop(self):
        self.reactor.stop()

    def add_listener(self, listener):
        if self._subscription_manager is not None:
            self._subscription_manager.add_listener(listener)
        else:
            raise Exception("Subscription manager is not enabled for this instance")

    def request_async(self, endpoint_name, endpoint_call_options, callback, cancellation_event):
        if endpoint_call_options.method_string == "POST":
            self.headers['Content-type'] = "application/json"

        def async_request(endpoint_call_options, cancellation_event, callback):
            def manage_failures(failure):
                # Cancelled
                if failure.type == ConnectingCancelledError:
                    return
                elif failure.type == PubNubTwistedException:
                    callback(failure.value)
                else:
                    return failure

            def options_func():
                return endpoint_call_options

            request = self.request_deferred(options_func, cancellation_event)
            request.addCallbacks(callback, manage_failures)

        self.reactor.callLater(0, async_request, endpoint_call_options, cancellation_event, callback)

        return

    # REVIEW: cancellation_event doesn't used inside function
    def request_deferred(self, options_func, cancellation_event):
        options = options_func()
        reactor = self.reactor
        pnconn_pool = self.pnconn_pool
        headers = self.headers
        params_to_merge_in = {}

        if options.operation_type == PNOperationType.PNPublishOperation:
            params_to_merge_in['seqn'] = self._publish_sequence_manager.get_next_sequence()

        options.merge_params_in(params_to_merge_in)

        create_response = options.create_response
        create_status_response = options.create_status

        url = utils.build_url(self.config.scheme(), self.base_origin,
                              options.path, options.query_string)

        logger.debug("%s %s %s" % (options.method_string, url, options.data))

        def handler():
            agent = Agent(reactor, pool=pnconn_pool)

            if options.data is not None:
                body = FileBodyProducer(StringIO(options.data))
            else:
                body = None
            request = agent.request(
                options.method_string,
                url,
                Headers(headers),
                body)

            def received(response):
                finished = Deferred()
                response.deliverBody(PubNubResponse(finished, response.code))
                return finished

            def success(response, req_url, request):
                parsed_url = urlparse(req_url)
                query = parse_qs(parsed_url.query)
                uuid = None
                auth_key = None

                if 'uuid' in query and len(query['uuid']) > 0:
                    uuid = query['uuid'][0]

                if 'auth_key' in query and len(query['auth_key']) > 0:
                    auth_key = query['auth_key'][0]

                response_body = response.body
                code = response.code
                d = Deferred()

                response_info = ResponseInfo(
                    status_code=response.code,
                    tls_enabled='https' == parsed_url.scheme,
                    origin=parsed_url.netloc,
                    uuid=uuid,
                    auth_key=auth_key,
                    client_request=request
                )

                if code != 200:
                    if code == 403:
                        status_category = PNStatusCategory.PNAccessDeniedCategory
                    elif code == 400:
                        status_category = PNStatusCategory.PNBadRequestCategory
                    else:
                        status_category = self

                    if code >= 500:
                        error = PNERR_SERVER_ERROR
                    else:
                        error = PNERR_CLIENT_ERROR
                else:
                    error = None
                    status_category = PNStatusCategory.PNAcknowledgmentCategory

                try:
                    data = json.loads(response_body)
                except ValueError:
                    try:
                        data = json.loads(response_body.decode("utf-8"))
                    except ValueError:
                        raise PubNubTwistedException(
                            result=create_response(None),
                            status=create_status_response(
                                status_category,
                                response_info,
                                PubNubException(
                                    pn_error=PNERR_JSON_DECODING_FAILED,
                                    errormsg='json decode error'
                                )
                            )
                        )

                if error:
                    raise PubNubTwistedException(
                        result=data,
                        status=create_status_response(status_category, data, response_info,
                                                      PubNubException(
                                                          errormsg=data,
                                                          pn_error=error,
                                                          status_code=response.code
                                                      )))

                envelope = TwistedEnvelope(
                    create_response(data),
                    create_status_response(
                        status_category,
                        response,
                        response_info,
                        error),
                    data
                )
                d.callback(envelope)
                return d

            def failed(failure):
                raise PubNubTwistedException(
                    result=None,
                    status=create_status_response(PNStatusCategory.PNTLSConnectionFailedCategory,
                                                  None,
                                                  None,
                                                  PubNubException(
                                                      errormsg=str(failure),
                                                      pn_error=PNERR_CONNECTION_ERROR,
                                                      status_code=0
                                                  )))
            request.addErrback(failed)
            request.addCallback(received)
            request.addCallback(success, url, request)

            return request

        return handler()

    def disconnected(self):
        return self.disconnected_times > 0


class TwistedEnvelope(object):
    def __init__(self, result, status, raw_result=None):
        self.result = result
        self.status = status
        self.raw_result = raw_result


class TwistedResponse(object):
    def __init__(self, body, code):
        self.body = body
        self.code = code


class PubNubTwistedException(Exception):
    def __init__(self, result, status):
        self.result = result
        self.status = status

    def __str__(self):
        return str(self.status.error_data.exception)



import json
import logging
import socket
import threading

import six
from six.moves import urllib

from pubnub import utils
from pubnub.enums import PNStatusCategory
from pubnub.errors import PNERR_CLIENT_ERROR, PNERR_UNKNOWN_ERROR, PNERR_CLIENT_TIMEOUT, \
    PNERR_HTTP_ERROR, PNERR_CONNECTION_ERROR
from pubnub.errors import PNERR_SERVER_ERROR
from pubnub.exceptions import PubNubException
from pubnub.request_handlers.base import BaseRequestHandler
from pubnub.structures import RequestOptions, PlatformOptions, ResponseInfo, Envelope

logger = logging.getLogger("pubnub")


class Urllib2RequestHandler(BaseRequestHandler):
    """
     PubNub Python SDK Native requests handler based on `urllib2/urllib` native HTTP library.

     Do not use this helper since it's doesnt finished yet. Treat it as an example how to write a custom
     handler for PubNub SDK
     """
    ENDPOINT_THREAD_COUNTER = 0

    def __init__(self, pubnub):
        self.pubnub = pubnub

    def sync_request(self, platform_options, endpoint_call_options):
        return self._build_envelope(platform_options, endpoint_call_options)

    def async_request(self, endpoint_name, platform_options, endpoint_call_options, callback, cancellation_event):
        call = Call()

        def callback_to_invoke_in_another_thread():
            try:
                envelope = self._build_envelope(platform_options, endpoint_call_options)
                if cancellation_event is not None and cancellation_event.isSet():
                    # Since there are no way to affect on ongoing request it's response will
                    # be just ignored on cancel call
                    return

                callback(envelope)
            except PubNubException as e:
                logger.error("Async request PubNubException. %s" % str(e))
                callback(Envelope(
                    result=None,
                    status=endpoint_call_options.create_status(
                        category=PNStatusCategory.PNBadRequestCategory,
                        response=None,
                        response_info=None,
                        exception=e)))
            except Exception as e:
                logger.error("Async request Exception. %s" % str(e))
                callback(Envelope(
                    result=None,
                    status=endpoint_call_options.create_status(
                        category=PNStatusCategory.PNInternalExceptionCategory,
                        response=None,
                        response_info=None,
                        exception=e)))
            finally:
                call.executed_cb()

        client = AsyncHTTPClient(callback_to_invoke_in_another_thread)

        thread = threading.Thread(
            target=client.run,
            name="EndpointThread-%s-%d" % (endpoint_name, ++Urllib2RequestHandler.ENDPOINT_THREAD_COUNTER)
        )
        thread.setDaemon(self.pubnub.config.daemon)
        thread.start()

        call.thread = thread
        call.cancellation_event = cancellation_event

        return call

    def _build_envelope(self, p_options, e_options):
        """ A wrapper for _invoke_url to separate request logic """

        status_category = PNStatusCategory.PNUnknownCategory
        response_info = None

        try:
            res = self._invoke_request(p_options, e_options, self.pubnub.base_origin)
        except PubNubException as e:
            if e._pn_error is PNERR_CONNECTION_ERROR:
                status_category = PNStatusCategory.PNUnexpectedDisconnectCategory
            elif e._pn_error is PNERR_CLIENT_TIMEOUT:
                status_category = PNStatusCategory.PNTimeoutCategory

            return Envelope(
                result=None,
                status=e_options.create_status(
                    category=status_category,
                    response=None,
                    response_info=response_info,
                    exception=e))

        if res is not None:
            url = six.moves.urllib.parse.urlparse(res.url)
            query = six.moves.urllib.parse.parse_qs(url.query)
            uuid = None
            auth_key = None

            if 'uuid' in query and len(query['uuid']) > 0:
                uuid = query['uuid'][0]

            if 'auth_key' in query and len(query['auth_key']) > 0:
                auth_key = query['auth_key'][0]

            response_info = ResponseInfo(
                status_code=res.code,
                tls_enabled='https' == url.scheme,
                origin=url.hostname,
                uuid=uuid,
                auth_key=auth_key,
                client_request=None
            )

        decoded_text = res.read().decode('utf-8')
        decoded_json = json.loads(decoded_text)
        logger.debug("GOT %s" % decoded_text)

        if res.code != 200:
            if res.code == 403:
                status_category = PNStatusCategory.PNAccessDeniedCategory

            if res.code == 400:
                status_category = PNStatusCategory.PNBadRequestCategory

            if decoded_json is None:
                text = "N/A"
            else:
                text = decoded_json

            if res.status_code >= 500:
                err = PNERR_SERVER_ERROR
            else:
                err = PNERR_CLIENT_ERROR

            return Envelope(
                result=e_options.create_response(decoded_json),
                status=e_options.create_status(
                    category=status_category,
                    response=decoded_json,
                    response_info=response_info,
                    exception=PubNubException(
                        pn_error=err,
                        errormsg=text,
                        status_code=res.status_code
                    )))
        else:

            return Envelope(
                result=e_options.create_response(decoded_json),
                status=e_options.create_status(
                    category=PNStatusCategory.PNAcknowledgmentCategory,
                    response=decoded_json,
                    response_info=response_info,
                    exception=None))

    @staticmethod
    def _invoke_request(p_options, e_options, base_origin):
        assert isinstance(p_options, PlatformOptions)
        assert isinstance(e_options, RequestOptions)

        url = utils.build_url(p_options.pn_config.scheme(), base_origin,
                              e_options.path, e_options.query_string)

        args = {
            "method": e_options.method_string,
            'headers': p_options.headers,
            "url": url,
            'params': e_options.query_string,
            'timeout': (e_options.connect_timeout, e_options.request_timeout)
        }

        if e_options.is_post() or e_options.is_patch():
            args['data'] = e_options.data
            logger.debug("%s %s %s" % (e_options.method_string, url, e_options.data))
        else:
            logger.debug("%s %s" % (e_options.method_string, url))

        try:
            req = urllib.request.Request(url, e_options.data, p_options.headers)
            res = urllib.request.urlopen(req)
        except urllib.error.URLError as e:
            # For Python 2.6
            if isinstance(e.reason, socket.timeout):
                raise PubNubException(
                    pn_error=PNERR_CLIENT_TIMEOUT,
                    errormsg=str(e)
                )
            else:
                # TODO: wrap
                raise

        except urllib.error.HTTPError as e:
            raise PubNubException(
                pn_error=PNERR_HTTP_ERROR,
                errormsg=str(e)
            )
        except socket.timeout as e:
            raise PubNubException(
                pn_error=PNERR_CLIENT_TIMEOUT,
                errormsg=str(e)
            )
        except Exception as e:
            raise PubNubException(
                pn_error=PNERR_UNKNOWN_ERROR,
                errormsg=str(e)
            )

        return res


class AsyncHTTPClient:
    """A wrapper for threaded calls"""

    def __init__(self, callback_to_invoke):
        self._callback_to_invoke = callback_to_invoke

    def run(self):
        self._callback_to_invoke()


class Call(object):
    """
    A platform dependent representation of async PubNub method call
    """

    def __init__(self):
        self.thread = None
        self.cancellation_event = None
        self.is_executed = False
        self.is_canceled = False

    def cancel(self):
        """
        Set Event flag to stop thread on timeout. This will not stop thread immediately, it will stopped
          only after ongoing request will be finished
        :return: nothing
        """
        if self.cancellation_event is not None:
            self.cancellation_event.set()
        self.is_canceled = True

    def join(self):
        if isinstance(self.thread, threading.Thread):
            self.thread.join()

    def executed_cb(self):
        self.is_executed = True



class PNGetSpacesResult(object):
    def __init__(self, result):
        """
        Representation of get spaces server response

        :param result: result of get spaces operation
        """
        self.data = result['data']
        self.status = result['status']
        self.total_count = result.get('totalCount', None)
        self.next = result.get('next', None)
        self.prev = result.get('prev', None)

    def __str__(self):
        return "Get spaces success with data: %s" % self.data


class PNCreateSpaceResult(object):
    def __init__(self, result):
        """
        Representation of create space server response

        :param result: result of create space operation
        """
        self.data = result['data']
        self.status = result['status']

    def __str__(self):
        return "Space created with data: %s" % self.data


class PNGetSpaceResult(object):
    def __init__(self, result):
        """
        Representation of get space server response

        :param result: result of get space operation
        """
        self.data = result['data']
        self.status = result['status']

    def __str__(self):
        return "Get space success with data: %s" % self.data


class PNUpdateSpaceResult(object):
    def __init__(self, result):
        """
        Representation of update space server response

        :param result: result of update space operation
        """
        self.data = result['data']
        self.status = result['status']

    def __str__(self):
        return "Update space success with data: %s" % self.data


class PNDeleteSpaceResult(object):
    def __init__(self, result):
        """
        Representation of delete space server response

        :param result: result of delete space operation
        """
        self.data = result['data']
        self.status = result['status']

    def __str__(self):
        return "Delete space success with data: %s" % self.data


class PNSpaceResult(object):
    def __init__(self, event, data):
        self.data = data
        self.event = event

    def __str__(self):
        return "Space %s event with data: %s" % (self.event, self.data)



import six

from pubnub.endpoints.endpoint import Endpoint
from pubnub.managers import TokenManagerProperties
from pubnub.models.consumer.user import PNGetUsersResult
from pubnub.enums import HttpMethod, PNOperationType, PNResourceType
from pubnub import utils


class GetUsers(Endpoint):
    GET_USERS_PATH = '/v1/objects/%s/users'
    MAX_LIMIT = 100

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._start = None
        self._end = None
        self._limit = GetUsers.MAX_LIMIT
        self._count = False
        self._include = None
        self._filter = None

    def start(self, start):
        assert isinstance(start, six.string_types)
        self._start = start
        return self

    def end(self, end):
        assert isinstance(end, six.string_types)
        self._end = end
        return self

    def limit(self, limit):
        assert isinstance(limit, six.integer_types)
        self._limit = limit
        return self

    def count(self, count):
        self._count = bool(count)
        return self

    def include(self, data):
        self._include = data
        return self

    def filter(self, _filter):
        assert isinstance(_filter, six.string_types)
        self._filter = _filter
        return self

    def custom_params(self):
        params = {}

        if self._start is not None:
            params['start'] = self._start

        if self._end is not None and self._start is None:
            params['end'] = self._end

        if self._count is True:
            params['count'] = True

        if self._limit != GetUsers.MAX_LIMIT:
            params['limit'] = self._limit

        if self._include:
            params['include'] = self._include

        if self._filter:
            params['filter'] = utils.url_encode(self._filter)

        return params

    def build_path(self):
        return GetUsers.GET_USERS_PATH % (self.pubnub.config.subscribe_key)

    def http_method(self):
        return HttpMethod.GET

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()

    def create_response(self, envelope):   # pylint: disable=W0221
        return PNGetUsersResult(envelope)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNGetUsersOperation

    def name(self):
        return 'Get users'

    def get_tms_properties(self):
        return TokenManagerProperties(
            resource_type=PNResourceType.USER,
            resource_id=""
        )
